""" Serializers used by the sample app """
from protean_flask.core.serializers import EntitySerializer
from protean_flask.core.serializers import ma

from .entities import Dog
from .entities import Human
from .entities import RelatedDog


class DogSerializer(EntitySerializer):
    """ Serializer for Dog Entity"""

    class Meta:
        entity = Dog


class HumanSerializer(EntitySerializer):
    """ Serializer for Human Entity"""

    class Meta:
        entity = Human


class HumanDetailSerializer(EntitySerializer):
    """ Serializer for the Human Entity with association"""
    dogs = ma.fields.Nested('RelatedDogSerializer', many=True,
                            exclude=['owner'])

    class Meta:
        entity = Human


class RelatedDogSerializer(EntitySerializer):
    """ Serializer for the Related Dpg Entity"""
    owner = ma.fields.Nested(HumanSerializer)

    class Meta:
        entity = RelatedDog



"""Module to setup Factories and other required artifacts for tests"""

import os

import mock
import pytest

os.environ['PROTEAN_CONFIG'] = 'tests.support.sample_config'


def pytest_addoption(parser):
    """Additional options for running tests with pytest"""
    parser.addoption(
        "--slow", action="store_true", default=False, help="run slow tests"
    )


def pytest_collection_modifyitems(config, items):
    """Configure special markers on tests, so as to control execution"""
    if config.getoption("--slow"):
        # --slow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --slow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


@pytest.fixture(scope="session", autouse=True)
def register_models():
    """Register Test Models with Dict Repo

       Run only once for the entire test suite
    """
    from protean.core.repository import repo_factory
    from tests.support.sample_app.entities import Dog, RelatedDog, Human

    repo_factory.register(Dog)
    repo_factory.register(RelatedDog)
    repo_factory.register(Human)


@pytest.fixture(autouse=True)
def run_around_tests():
    """Cleanup Database after each test run"""
    from protean.core.repository import repo_factory
    from tests.support.sample_app.entities import Dog, RelatedDog, Human

    # A test function will be run at this point
    yield

    repo_factory.get_repository(Dog).delete_all()
    repo_factory.get_repository(RelatedDog).delete_all()
    repo_factory.get_repository(Human).delete_all()



""" Usecases of the sample application """

from protean.core.tasklet import Tasklet
from protean.core.usecase import ListRequestObject
from protean.core.usecase import ListUseCase
from protean.core.usecase import ShowRequestObject
from protean.core.usecase import UseCase

from .entities import Dog
from .entities import Human


class ListMyDogsRequestObject(ShowRequestObject):
    """ Request object for the my dogs use case """


class ListMyDogsUsecase(UseCase):
    """ Use case for listing dogs of a human """
    def process_request(self, request_object):
        """ Get the human and return the dogs owned by the human"""
        human = Human.get(request_object.identifier)

        # Get the dogs related to the human
        payload = {'owner': human.name, 'order_by': ['age']}
        response_object = Tasklet.perform(
            Dog, ListUseCase, ListRequestObject, payload, raise_error=True)
        return response_object



""" Blueprints of the sample app"""
from flask import Blueprint
from tests.support.sample_app.entities import Dog
from tests.support.sample_app.entities import Human
from tests.support.sample_app.serializers import DogSerializer
from tests.support.sample_app.serializers import HumanSerializer
from tests.support.sample_app.usecases import ListMyDogsRequestObject
from tests.support.sample_app.usecases import ListMyDogsUsecase

from protean_flask import Protean
from protean_flask.core.views import ShowAPIResource
from protean_flask.core.viewsets import GenericAPIResourceSet

blueprint = Blueprint('test_blueprint', __name__)
api_bp = Protean(blueprint)


class ShowDogResource(ShowAPIResource):
    """ View for retrieving a Dog by its ID"""
    entity_cls = Dog
    serializer_cls = DogSerializer


class HumanResourceSet(GenericAPIResourceSet):
    """ Resource Set for the Human Entity"""
    entity_cls = Human
    serializer_cls = HumanSerializer

    def my_dogs(self, identifier):
        """ List all the dogs belonging to the Human"""
        # Run the usecase and get the related dogs
        payload = {'identifier': identifier}
        response_object = self._process_request(
            ListMyDogsUsecase, ListMyDogsRequestObject, payload=payload,
            no_serialization=True)

        # Serialize the results and return the response
        serializer = DogSerializer(many=True)
        items = serializer.dump(response_object.value.items)
        result = {
            'dogs': items.data,
            'total': response_object.value.total,
            'page': response_object.value.page
        }
        return result, response_object.code.value


# Setup the routes
blueprint.add_url_rule(
    '/dogs/<int:identifier>', view_func=ShowDogResource.as_view('show_dog'),
    methods=['GET'])
api_bp.register_viewset(
    HumanResourceSet, 'humans', '/humans', pk_type='int',
    additional_routes=['/<int:identifier>/my_dogs'])



"""Module to test Serializer functionality"""

import marshmallow as ma
import pytest
from protean.core.exceptions import ConfigurationError

from protean_flask.core.serializers import EntitySerializer

from ..support.sample_app.entities import Dog
from ..support.sample_app.entities import Human
from ..support.sample_app.entities import RelatedDog
from ..support.sample_app.serializers import HumanDetailSerializer
from ..support.sample_app.serializers import RelatedDogSerializer


class DogSerializer(EntitySerializer):
    """ Serializer for the Dog Entity """
    class Meta:
        entity = Dog


class TestEntitySerializer:
    """Tests for EntitySerializer class"""

    def test_init(self):
        """Test initialization of EntitySerializer derived class"""
        s = DogSerializer()
        assert s is not None

        # Check that the entity gets serialized correctly
        s_result = s.dump(Dog(id=1, name='Johnny', owner='John'))
        expected_data = {'age': 5, 'id': 1, 'name': 'Johnny', 'owner': 'John'}
        assert s_result.data == expected_data

    def test_abstraction(self):
        """Test that EntitySerializer class itself cannot be initialized"""

        with pytest.raises(ConfigurationError):
            EntitySerializer()

    def test_include_fields(self):
        """ Test the include fields option of the serializer"""

        class DogSerializer2(EntitySerializer):
            """ Serializer for the Dog Entity """
            class Meta:
                entity = Dog
                fields = ('id', 'age')

        s = DogSerializer2()
        assert s is not None

        # Check that the entity gets serialized correctly
        s_result = s.dump(Dog(id=1, name='Johnny', owner='John'))
        expected_data = {'age': 5, 'id': 1}
        assert s_result.data == expected_data

    def test_exclude_fields(self):
        """ Test the exclude fields option of the serializer"""

        class DogSerializer2(EntitySerializer):
            """ Serializer for the Dog Entity """
            class Meta:
                entity = Dog
                exclude = ('id', 'age')

        s = DogSerializer2()
        assert s is not None

        # Check that the entity gets serialized correctly
        s_result = s.dump(Dog(id=1, name='Johnny', owner='John'))
        expected_data = {'name': 'Johnny', 'owner': 'John'}
        assert s_result.data == expected_data

    def test_method_fields(self):
        """ Test the method field type of the serializer"""

        class DogSerializer2(EntitySerializer):
            """ Serializer for the Dog Entity """
            old = ma.fields.Method('get_old')

            def get_old(self, obj):
                """ Check if the dog is old or young """
                if obj.age > 5:
                    return True
                else:
                    return False

            class Meta:
                entity = Dog

        s = DogSerializer2()
        assert s is not None

        # Check that the entity gets serialized correctly
        s_result = s.dump(Dog(id=1, name='Johnny', owner='John'))
        expected_data = {
            'name': 'Johnny',
            'owner': 'John',
            'age': 5,
            'id': 1,
            'old': False
        }
        assert s_result.data == expected_data


class TestEntitySerializer2:
    """Tests for EntitySerializer class with related fields """

    @classmethod
    def setup_class(cls):
        """ Setup the test case """
        cls.human = Human.create(id=1, name='John')

    def test_reference_field(self):
        """ Test that the reference field gets serialized """

        dog = RelatedDog.create(id=5, name='Johnny', owner=self.human)

        # Check that the entity gets serialized correctly
        s = RelatedDogSerializer()
        s_result = s.dump(dog)
        expected_data = {
            'name': 'Johnny',
            'owner':  {'contact': None, 'id': 1, 'name': 'John'},
            'age': 5,
            'id': 5,
        }
        assert s_result.data == expected_data

    def test_hasmany_association(self):
        """ Test the has many association gets serialized """
        RelatedDog.create(id=5, name='Johnny', owner=self.human)

        s = HumanDetailSerializer()
        s_result = s.dump(self.human)
        expected_data = {
            'name': 'John',
            'dogs': [{'age': 5, 'id': 5, 'name': 'Johnny'}],
            'id': 1,
            'contact': None
        }
        assert s_result.data == expected_data



""" Views of the sample app"""
from protean.context import context

from protean_flask.core.views import APIResource
from protean_flask.core.views import CreateAPIResource
from protean_flask.core.views import DeleteAPIResource
from protean_flask.core.views import ListAPIResource
from protean_flask.core.views import ShowAPIResource
from protean_flask.core.views import UpdateAPIResource
from protean_flask.core.viewsets import GenericAPIResourceSet

from .entities import Dog
from .entities import Human
from .serializers import DogSerializer
from .serializers import HumanSerializer
from .usecases import ListMyDogsRequestObject
from .usecases import ListMyDogsUsecase


class ShowDogResource(ShowAPIResource):
    """ View for retrieving a Dog by its ID"""
    entity_cls = Dog
    serializer_cls = DogSerializer


class ListDogResource(ListAPIResource):
    """ View for listing Dog entities"""
    entity_cls = Dog
    serializer_cls = DogSerializer


class CreateDogResource(CreateAPIResource):
    """ View for creating a Dog Entity"""
    entity_cls = Dog
    serializer_cls = DogSerializer


class UpdateDogResource(UpdateAPIResource):
    """ View for updating a Dog by its ID"""
    entity_cls = Dog
    serializer_cls = DogSerializer


class DeleteDogResource(DeleteAPIResource):
    """ View for deleting a Dog by its ID"""
    entity_cls = Dog
    serializer_cls = DogSerializer


class HumanResourceSet(GenericAPIResourceSet):
    """ Resource Set for the Human Entity"""
    entity_cls = Human
    serializer_cls = HumanSerializer

    def my_dogs(self, identifier):
        """ List all the dogs belonging to the Human"""
        # Run the usecase and get the related dogs
        payload = {'identifier': identifier}
        dogs_list = self._process_request(
            ListMyDogsUsecase, ListMyDogsRequestObject, payload=payload,
            no_serialization=True)

        # Serialize the results and return the response
        serializer = DogSerializer(many=True)
        items = serializer.dump(dogs_list.items)
        page = int(dogs_list.offset / dogs_list.limit) + 1
        result = {
            'dogs': items.data,
            'total': dogs_list.total,
            'page': page
        }
        return result, 200


def flask_view():
    """ A non protean flask view """
    return 'View Response', 200


class CurrentContextResource(APIResource):
    """ View for retrieving the current context information """

    def get(self):
        """ Return the context information on GET """
        context_data = {
            'host_url': context.host_url,
            'url': context.url,
            'tenant_id': context.tenant_id,
            'user_agent': context.user_agent,
            'user_agent_hash': context.user_agent_hash,
            'remote_addr': context.remote_addr
        }
        return context_data



"""This module exposes the Base Resource View for all Application Views"""

import inflect
from flask import Response
from flask import current_app
from flask import request
from flask.views import MethodView
from protean.conf import active_config
from protean.core.tasklet import Tasklet
from protean.core.transport import Status
from protean.core.usecase import CreateRequestObject
from protean.core.usecase import CreateUseCase
from protean.core.usecase import DeleteRequestObject
from protean.core.usecase import DeleteUseCase
from protean.core.usecase import ListRequestObject
from protean.core.usecase import ListUseCase
from protean.core.usecase import ShowRequestObject
from protean.core.usecase import ShowUseCase
from protean.core.usecase import UpdateRequestObject
from protean.core.usecase import UpdateUseCase
from protean.utils import inflection
from protean.utils.importlib import perform_import
from werkzeug.http import parse_options_header

from protean_flask.utils import immutable_dict_2_dict

INFLECTOR = inflect.engine()


class APIResource(MethodView):
    """The base resource view that allows defining custom methods other than
    the standard five REST routes. Also handles rendering the output to json.

    """

    def _lookup_method(self):
        """ Lookup the class method to be called for this request"""
        func = request.url_rule.rule.rsplit('/', 1)[-1]
        meth = None
        if func and (func[0] != '<' and func[1] != '>'):
            meth = getattr(self, func, None)

        if not meth:
            meth = getattr(self, request.method.lower(), None)

        # If the request method is HEAD and we don't have a handler for it
        # retry with GET.
        if meth is None and request.method == 'HEAD':
            meth = getattr(self, 'get', None)

        assert meth is not None, 'Unimplemented method %r' % request.method
        return meth

    def parse_payload(self):
        """ Load the data from form, json and query to `request.payload` """

        # Parse the request content based on the content type
        content_type = (request.content_type or
                        active_config.DEFAULT_CONTENT_TYPE)
        mime_type, _ = parse_options_header(content_type)

        if request.method in ['POST', 'PUT', 'PATCH']:
            if mime_type == 'application/x-www-form-urlencoded':
                request.payload = immutable_dict_2_dict(request.form)
            elif mime_type == 'multipart/form-data':
                request.payload = immutable_dict_2_dict(request.form)
                request.payload.update(
                    immutable_dict_2_dict(request.files))
            elif mime_type == 'application/json':
                request.payload = request.get_json(
                    force=True, silent=True) or {}

        elif request.method == 'GET':
            request.payload = immutable_dict_2_dict(request.args)

        # If a customer parser is defined then run that
        parser = getattr(self, 'parser', None)
        if parser:
            parser = perform_import(parser)
            parser()

    def render_response(self, response):
        """ Render the response to the expected format """

        # Get the renderer to be used for this view
        renderer = getattr(self, 'renderer', None)
        if not renderer:
            renderer = perform_import(active_config.DEFAULT_RENDERER)

        # Perform rendering only for non Flask Responses
        if not isinstance(response, current_app.response_class):
            data, code, headers = self._unpack_response(response)
            response = renderer(data, code, headers)

        return response

    @staticmethod
    def _unpack_response(value):
        """Return a three tuple of data, code, and headers"""
        if not isinstance(value, tuple):
            return value, 200, {}

        try:
            data, code, headers = value
            return data, code, headers
        except ValueError:
            pass

        try:
            data, code = value
            return data, code, {}
        except ValueError:
            pass

        return value, 200, {}

    def dispatch_request(self, *args, **kwargs):
        """Dispatch the request to the correct function"""

        # Lookup method defined for this resource
        meth = self._lookup_method()

        # Parse the request input and populate the payload
        self.parse_payload()

        # Call the method and create the response
        response = meth(*args, **kwargs)
        final_response = self.render_response(response)
        return final_response


class GenericAPIResource(APIResource):
    """This is the Generic Base Class for all Views
    """

    entity_cls = None
    serializer_cls = None

    def get_entity_cls(self):
        """
        Return the class to use for the serializer.
        Defaults to using `self.schema_cls`.
        """
        if not self.entity_cls:
            raise AssertionError(
                "'%s' should either include a `entity_cls` attribute, "
                "or override the `get_entity_cls()` method."
                % self.__class__.__name__
            )
        return self.entity_cls

    def get_serializer(self, *args, **kwargs):
        """
        Return the serializer instance that should be used for validating and
        de-serializing input, and for serializing output.
        """
        serializer_cls = self.get_serializer_cls()
        serializer = serializer_cls(*args, **kwargs)
        serializer.context = self.get_serializer_context()
        return serializer

    def get_serializer_cls(self):
        """
        Return the class to use for the serializer.
        Defaults to using `self.serializer_class`.
        """
        if not self.serializer_cls:
            raise AssertionError(
                "'%s' should either include a `serializer_cls` attribute, "
                "or override the `get_serializer_cls()` method."
                % self.__class__.__name__
            )
        return self.serializer_cls

    def get_serializer_context(self):
        """
        Extra context provided to the serializer class.
        """
        return {
            'view': self
        }

    def _process_request(self, usecase_cls, request_object_cls, payload,
                         many=False, no_serialization=False):
        """ Process the request by running the Protean Tasklet """
        # Get the schema class and derive resource name
        entity_cls = self.get_entity_cls()
        resource = inflection.underscore(entity_cls.__name__)

        # Get the serializer for this class
        serializer = None
        if not no_serialization:
            serializer = self.get_serializer(many=many)

        # Run the use case and return the results
        response_object = Tasklet.perform(
            entity_cls, usecase_cls, request_object_cls, payload,
            raise_error=True)

        # If no serialization is set just return the response object
        if no_serialization:
            return response_object.value

        # Return empty response for 204s
        if response_object.code == Status.SUCCESS_WITH_NO_CONTENT:
            return Response(None, response_object.code.value)

        # Serialize the results and return the response
        if many:
            items = serializer.dump(response_object.value.items)
            page = int(response_object.value.offset / response_object.value.limit) + 1
            result = {
                INFLECTOR.plural(resource): items.data,
                'total': response_object.value.total,
                'page': page
            }
            return result, response_object.code.value

        else:
            result = serializer.dump(response_object.value)
            return {resource: result.data}, response_object.code.value


class ShowAPIResource(GenericAPIResource):
    """ An API view for retrieving an entity by its identifier"""

    usecase_cls = ShowUseCase
    request_object_cls = ShowRequestObject

    def get(self, identifier):
        """Show the entity.
        Expected Parameters:
             identifier = <string>, identifies the entity
        """
        payload = {'identifier': identifier}
        return self._process_request(
            self.usecase_cls, self.request_object_cls, payload=payload)


class ListAPIResource(GenericAPIResource):
    """ An API view for listing entities"""

    usecase_cls = ListUseCase
    request_object_cls = ListRequestObject

    def get(self):
        """List the entities.
        """
        return self._process_request(
            self.usecase_cls, self.request_object_cls,
            payload=request.payload, many=True)


class CreateAPIResource(GenericAPIResource):
    """ An API view for creating an entity"""

    usecase_cls = CreateUseCase
    request_object_cls = CreateRequestObject

    def post(self):
        """Create the entity.
        """
        return self._process_request(
            self.usecase_cls, self.request_object_cls, payload=request.payload)


class UpdateAPIResource(GenericAPIResource):
    """ An API view for updating an entity"""

    usecase_cls = UpdateUseCase
    request_object_cls = UpdateRequestObject

    def put(self, identifier):
        """Update the entity.
         Expected Parameters:
             identifier = <string>, identifies the entity
        """
        payload = {
            'identifier': identifier,
            'data': request.payload
        }
        return self._process_request(
            self.usecase_cls, self.request_object_cls, payload=payload)


class DeleteAPIResource(GenericAPIResource):
    """ An API view for deleting an entity"""

    usecase_cls = DeleteUseCase
    request_object_cls = DeleteRequestObject

    def delete(self, identifier):
        """Delete the entity.
         Expected Parameters:
             identifier = <string>, identifies the entity
        """
        payload = {'identifier': identifier}
        return self._process_request(
            self.usecase_cls, self.request_object_cls, payload=payload)



"""This module holds the generic definitions of Serializer"""

from collections import OrderedDict

import marshmallow as ma
from protean.core import field
from protean.core.entity import Entity
from protean.core.exceptions import ConfigurationError


class BaseSerializer(ma.Schema):
    """Base serializer with which to define custom serializers."""


class EntitySerializerOpts(ma.schema.SchemaOpts):
    """ Options for the entity serializer"""
    def __init__(self, meta):
        super().__init__(meta)
        self.entity_cls = getattr(meta, 'entity', None)


class EntitySerializer(BaseSerializer):
    """Serializer which uses Entity class to automatically infer fields."""

    OPTIONS_CLASS = EntitySerializerOpts

    field_mapping = {
        field.String: ma.fields.String,
        field.Boolean: ma.fields.Boolean,
        field.Integer: ma.fields.Integer,
        field.Float: ma.fields.Float,
        field.List: ma.fields.List,
        field.Dict: ma.fields.Dict
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Updates the declared fields with the fields of the Entity class
        if not self.opts.entity_cls or not \
                issubclass(self.opts.entity_cls, Entity):
            raise ConfigurationError(
                '`Meta.entity` option must be set and a subclass of `Entity`.')

        entity_fields = OrderedDict()
        for field_name, field_obj in \
                self.opts.entity_cls.meta_.declared_fields.items():
            if self.opts.fields and field_name not in self.opts.fields:
                continue
            elif self.opts.exclude and field_name in self.opts.exclude:
                continue
            elif isinstance(field_obj, field.Reference):
                continue
            elif field_name not in self.declared_fields:
                entity_fields[field_name] = self.build_field(field_obj)

        self.declared_fields.update(entity_fields)

    def build_field(self, field_obj):
        """ Map the Entity field to a Marshmallow field """

        # Lookup the field mapping in the dictionary, default to String field
        e_field_type = type(field_obj)
        if e_field_type in self.field_mapping:
            field_opts = {}
            if e_field_type == field.List:
                field_opts['cls_or_instance'] = ma.fields.String
            return self.field_mapping[e_field_type](**field_opts)
        elif e_field_type == field.Auto:
            raise ConfigurationError(
                'Fields of type `Auto` must be manually set in the serializer.')
        else:
            return ma.fields.String()



"""Module to test View functionality and features"""

from tests.support.sample_app import app
from tests.support.sample_app.entities import Dog
from tests.support.sample_app.entities import Human


class TestBlueprint:
    """Class to test Blueprint functionality of flask with this package"""

    @classmethod
    def setup_class(cls):
        """ Setup for this test case"""

        # Create the test client
        cls.client = app.test_client()

    def test_show(self):
        """ Test retrieving an entity using blueprint ShowAPIResource"""

        # Create a dog object
        Dog.create(id=5, name='Johnny', owner='John')

        # Fetch this dog by ID
        rv = self.client.get('/blueprint/dogs/5')
        assert rv.status_code == 200

        expected_resp = {
            'dog': {'age': 5, 'id': 5, 'name': 'Johnny', 'owner': 'John'}
        }
        assert rv.json == expected_resp

        # Test search by invalid id
        rv = self.client.get('/blueprint/dogs/6')
        assert rv.status_code == 404

        # Delete the dog now
        dog = Dog.get(5)
        dog.delete()

    def test_set_show(self):
        """ Test retrieving an entity using the blueprint resource set"""
        # Create a human object
        Human.create(id=1, name='John')

        # Fetch this human by ID
        rv = self.client.get('/blueprint/humans/1')
        assert rv.status_code == 200
        expected_resp = {
            'human': {'contact': None, 'id': 1, 'name': 'John'}
        }
        assert rv.json == expected_resp

        # Delete the human now
        human = Human.get(1)
        human.delete()

    def test_custom_route(self):
        """ Test custom routes using the blueprint resource set """

        # Create a human object
        Human.create(id=1, name='John')
        Dog.create(id=5, name='Johnny', owner='John')

        # Get the custom route
        rv = self.client.get('/humans/1/my_dogs')
        assert rv.status_code == 200
        assert rv.json['total'] == 1
        assert rv.json['dogs'][0] == {'age': 5, 'id': 5, 'name': 'Johnny', 'owner': 'John'}



from protean.conf import active_config

from . import config
from .core.base import Protean

__version__ = '0.0.11'

__all__ = ('Protean',)


# Update the config here so that loading the repo will load the config
active_config.update_defaults(config)



""" Configuration variables used by Protean-Flask """

# Default output renderer for the app
DEFAULT_RENDERER = 'protean_flask.core.renderers.render_json'

# Default content type of the input data if none provided
DEFAULT_CONTENT_TYPE = 'application/json'

# Custom exception handler for the app
EXCEPTION_HANDLER = None



""" Utility functions used by Protean Flask"""


def immutable_dict_2_dict(imm_dict):
    """ Function to convert an Immutable Dictionary to a Mutable one
    Convert multi valued and keys ending with [] to lists
    """
    m_dict = {}

    for key, val in imm_dict.to_dict(flat=False).items():
        if len(val) > 1 or key.endswith('[]'):
            m_dict[key.strip('[]')] = val
        else:
            m_dict[key] = val[0]

    return m_dict


def derive_tenant(url):
    """Derive tenant ID from host

    We consider the standard `subdomain.domain.com` structure to be the
    `tenant_id`.

    There might be multiple applications that are hosted for the subdomain,
    and they may have additional application identifiers at the beginning,
    like 'customers.subdomain.domain.com' or 'backoffice.subdomain.domain.com'.
    In all such cases, we still consider the 3 part structure,
     `subdomain.domain.com`, to be the `tenant_id`.

    """

    from urllib.parse import urlparse

    host = urlparse(url).hostname

    return host



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""Setup Module for Protean Flask Extension"""

from __future__ import absolute_import
from __future__ import print_function

import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    """Helper method to read files"""
    return io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ).read()


setup(
    name='protean-flask',
    version='0.0.11',
    license='BSD 3-Clause License',
    description='Protean Flask Extension',
    long_description='%s\n%s' % (
        re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', read('README.rst')),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))
    ),
    author='Subhash Bhushan C',
    author_email='subhash@team8solutions.com',
    url='https://github.com/proteanhq/protean-flask',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 2 - Pre-Alpha',
        'Framework :: Flask',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: Unix',
        'Operating System :: POSIX :: Linux',
        'Operating System :: OS/2',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Database',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    install_requires=[
        'click==7.0',
        'werkzeug==0.15.2',
        'flask==1.0.2',
        'inflect==1.0.1',
        'protean==0.0.11',
        'marshmallow==2.16.1',
        # eg: 'aspectlib==1.1.1', 'six>=1.7',
    ],
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
    },
    entry_points={
        'console_scripts': [
            'protean-flask = protean_flask.cli:main',
        ]
    },
)



""" Entities used by the sample application """

from protean.core import field
from protean.core.entity import Entity
from protean.core.field import association


class Dog(Entity):
    """This is a dummy Dog Entity class"""
    id = field.Integer(identifier=True)
    name = field.String(required=True, max_length=50)
    age = field.Integer(default=5)
    owner = field.String(required=True, max_length=15)


class Human(Entity):
    """ This is a dummy Human class """
    id = field.Integer(identifier=True)
    name = field.String(required=True, max_length=50)
    contact = field.StringMedium()
    dogs = association.HasMany('RelatedDog', via='owner_id')


class RelatedDog(Entity):
    """ This is a dummy Dog with related field"""
    id = field.Integer(identifier=True)
    name = field.String(required=True, max_length=50)
    age = field.Integer(default=5)
    owner = field.Reference(Human)
